# Copyright 2023, Samuel Muscinelli.
# Code to produce the main results of Muscinelli, Wagner, and Litwin-Kumar (2023).
#  This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import math
import numpy as np
import copy
import torch
import torch.nn as nn
from torch.distributions.multivariate_normal import MultivariateNormal
from tools import genhaarorthmatrix, genclustembeddingmatrix, init_recurrent_connectivity, runforward, runandlearn, get_noise

# parse arguments
parser = argparse.ArgumentParser(
    description='Computes the performance on a random classification task of random, PC-aligned, whitening, and learned compression, as well as of single-step expansion.',
    epilog=''
)
parser.add_argument('--embedding', metavar='embedding', type=str, default="distributed", help='Type of embedding. Can be "clustered-uncorrelated", "clustered-correlated", or "distributed".')
parser.add_argument('--loss_type', metavar='loss_type', type=str, default="performance", help='Loss type for learned compression. Can be "performance" or "SNR_c".')
parser.add_argument('-r', '--rec', dest='recInh', action='store_true', help='Whether to have recurrent interactions.')
parser.add_argument('--rec_type', metavar='rec_type', type=str, default="learned", help='Type of recurrent interactions (if rec == True). Can be "learned", "global", or "unconstrained".')
parser.add_argument('--exc_only', dest='excOnly', action='store_true', help='whether to use only excitatory compression weights')
parser.add_argument('--relu', dest='relu_comp', action='store_true', help='whether to use nonlinear compression layer units.')

args = parser.parse_args()

#################
# Parse Arguments
#################

lossType = args.loss_type
embedding = args.embedding
recInhType = args.rec_type

# Boolean parameters
reluComp = args.relu_comp
excOnly = args.excOnly
recInh = args.recInh

# numerical parameters
D = 50
N = 500
sigmaO = 0.1
Nc = 50

task = "classification"
zStats = "manual"

# Dn = 100  # in most panels, this is equal to N
Dn = N  # in most panels, this is equal to N
M = 2000
K = 4
f = 0.1  # coding level ; 0.1 for most panels
fC = 0.3  # coding level of nonlinear compression
gI0 = 10.  # value (or initial value) of global inhibition strength

# task parameters
alpha = 1.0
P = int(alpha * D)
Pinf = 1000  # number of samples to study covariance structure
R = 10
Ptest = R * P
p = 1.  # eigenvalue decay speed ; 1 for most panels
# pn = 1.  # same as above, but for noise (in most panels this is 0)
pn = 0.
# p = 1.
sigmaP = 0.  # input noise strength
sigmaM = 0.  # noise strength at expansion layer

# learning rule parameters
readouttype = "Hebbian"
lrargs = {'meansubtracted': True}

# learning parameters
lr = 1e-3
nEpochs = 10000
B = 10  # number of batch samples
differentJperepoch = True

############################
# Generate fixed quantities
############################

# signal and noise distributions
SigmaZ = torch.diag(torch.pow(torch.arange(1, D + 1), -p))  # cov matrix of task variables
SigmaZn = torch.diag(torch.pow(torch.arange(1, Dn + 1), -pn))  # noise covariance matrix
distrZ = MultivariateNormal(torch.zeros(D), SigmaZ)
distrZn = MultivariateNormal(torch.zeros(Dn), SigmaZn)

# generate embedding matrix
if embedding not in ["clustered-uncorrelated", "clustered-correlated", "distributed"]:
    print("Warning! Embedding not recognized, falling back on distributed")
    embedding = "distributed"
if embedding == "distributed":
    Om = genhaarorthmatrix(N)
    A = torch.from_numpy(Om[:, :D]).float()  # this is the embedding matrix
    if Dn < N:
        # if noise is low-D, we assume it lives in the orthogonal complement
        An = torch.from_numpy(Om[:, D:D + Dn]).float()
    elif Dn == N:
        # is noise is high-D, we assume it is uncorrelated i.i.d. Gaussian
        An = torch.eye(Dn)
elif embedding == "clustered-uncorrelated":
    A = genclustembeddingmatrix(N, D)
    # We set noise to always be i.i.d Gaussian for clustered representations:
    An = torch.eye(N)
    Dn = N
    SigmaZn = torch.eye(N)
    pn = 0.
elif embedding == "clustered-correlated":
    Otmp = torch.from_numpy(genhaarorthmatrix(D)).float()
    A = genclustembeddingmatrix(N, D) @ Otmp  # the rotation via Otmp introduces correlations among clusters, provided that p > 0.
    # We set noise to always be i.i.d Gaussian for clustered representations:
    An = torch.eye(N)
    Dn = N
    SigmaZn = torch.eye(N)

###########################
# Define compression layers
###########################
cmprs = {
    "DE": nn.Linear(N, N, bias=False),  # fake compression that is used to implement the single-step expansion network
    "rnd": nn.Linear(N, Nc, bias=False),  # random compression
    "pca": nn.Linear(N, Nc, bias=False),  # PC-aligned compression
    "wtn": nn.Linear(N, Nc, bias=False),  # whitening compression
    "bkp": nn.Linear(N, Nc, bias=False)  # backprop compression
}
for key, c in cmprs.items():
    if key != "bkp":
        c.weight.requires_grad = False  # no learning if not Backprop

# set compression matrices
cmprs["DE"].weight.data = np.sqrt(N) * torch.eye(N)  # effectively no compression
cmprs["rnd"].weight.data = torch.randn(Nc, N)  # random compression (there is no scaling here because it is in the dynamics)
lambdas, U = torch.linalg.eigh(SigmaZ)  # eigendecomposition of SigmaZ, used to find the optimal compression matrix
Aeff = A @ U  # effective embedding matrix, i.e. product of the emedding matrix and the eigenvecotrs of SigmaZ

# define PC-aligned compression and whitening compression:
if Nc > D:  # if there are more comrpession layer units than task variables, repeat them
    i = 0
    while (i + 1) * D <= Nc:
        cmprs["pca"].weight.data[i * D:min((i + 1) * D, Nc), :] = np.sqrt(D) * Aeff[:, :].transpose(0, 1)
        for j in range(min(D, Nc - i * D)):
            cmprs["wtn"].weight.data[i * D + j, :] = copy.copy(cmprs["pca"].weight.data[j, :])
            cmprs["wtn"].weight.data[i * D + j, :] = cmprs["wtn"].weight.data[i * D + j, :] / math.sqrt(lambdas[j])
        i += 1
else:
    cmprs["pca"].weight.data[:, :] = np.sqrt(D) * Aeff[:, :Nc].transpose(0, 1)
    for j in range(Nc):
        cmprs["wtn"].weight.data[j, :] = copy.copy(cmprs["pca"].weight.data[j, :])
        cmprs["wtn"].weight.data[j, :] = cmprs["wtn"].weight.data[j, :] / math.sqrt(lambdas[j])

cmpr = cmprs["bkp"]
# initialize recurrent connectivity at the compression layer if necessary
if recInh:
    W = init_recurrent_connectivity(Nc, gI0, recInhType)
else:
    W = []

# train bkp compression using backprop
opt = torch.optim.Adam(cmpr.parameters(), lr=lr)  # optimizer
if recInh:
    if recInhType in ["learned", "unconstrained"]:
        opt = torch.optim.Adam(list(cmpr.parameters()) + [W], lr=lr)

losses = np.zeros(nEpochs)
frErrors = np.zeros(nEpochs)
for e in range(nEpochs):
    opt.zero_grad()
    zTrain = distrZ.sample(sample_shape=(P, B)).transpose(0, 2)  # sample task variables (training)
    zTest = zTrain + sigmaP * distrZ.sample(sample_shape=(P, B)).transpose(0, 2)  # test patterns: same as train + noise
    y = torch.from_numpy(np.random.choice([-1, 1], size=(B, P))).float()  # sample labels

    if differentJperepoch:
        # sample a new expansion matrix
        J = torch.zeros(M, Nc)
        for i in range(M):
            idxs = np.random.choice(np.arange(Nc), size=K, replace=False)
            J[i, idxs] = 1 / math.sqrt(K) * torch.randn(K)

    # run network forward, learn hebbian weights and test
    PyHat, _, c0, _, _ = runandlearn(cmpr, zTrain, zTest, y, math.sqrt(N / D) * A, J, An, distrZn, f=f, sigmaO=sigmaO, sigmaM=sigmaM, reluComp=reluComp, fC=fC, rec=recInh, W=W)

    # adjust compression weights using gradient descent
    if lossType not in ["performance", "SNR_c"]:
        print("Warning: lossType not recognized, falling back on 'performance'")
        lossType = "performance"
    if lossType == "performance":
        crit = nn.BCELoss()  # binary cross-entropy loss
        loss = crit(PyHat, (y + 1) / 2)
    elif lossType == "SNR_c":
        # use the SNR at the compression layer as the loss (used to train to approximate optimal compression)
        Cc = 1 / P * torch.einsum('ibp, jbp -> ijb', c0, c0)
        if Dn < N:
            Cn = An @ SigmaZn @ An.T
        else:
            Cn = torch.eye(N)
        dim = torch.sum(torch.diagonal(Cc, dim1=0, dim2=1), 1) ** 2 / torch.sum(torch.diagonal(torch.einsum('ijb, jkb -> ikb', Cc, Cc), dim1=0, dim2=1), 1)
        noise = get_noise(cmpr.weight, Cc, Cn, sigmaO)  # uses analytical expression AND assumes that isotropic noise is the only one
        loss = torch.mean(1 - dim * (torch.maximum(1 - noise, torch.zeros(noise.shape)))**2 / D)
    loss.backward()  # compute gradients

    opt.step()  # update compression weights

    if excOnly:
        cmpr.weight.data = torch.clamp(cmpr.weight, min=0., max=torch.inf)  # constrain weights to be positive by clipping
    if recInh and (recInhType != "unconstrained"):
        W.data = torch.clamp(W, min=-torch.inf, max=0.)

    if e % 10 == 0:
        print("Epoch: ", e, " ; Loss: ", loss)

    with torch.no_grad():
        # store loss and performance during training
        losses[e] = loss.detach().numpy()
        frErrors[e] = 1 / (2 * P * B) * np.sum(np.abs((PyHat > 0.5).detach().numpy() - ((y + 1) / 2).detach().numpy()), axis=(0, 1))

####################
# Test all networks
####################
# setup result dictionaries.
frErrorsAll = {}
dimCall = {}
dimAll = {}
noiseAll = {}
noiseCall = {}
noiseAllTh = {}

with torch.no_grad():

    zTrain = distrZ.sample(sample_shape=(P, B)).transpose(0, 2)  # sample training task subspace patterns
    zTest = torch.cat(10 * [zTrain], dim=2) + sigmaP * distrZ.sample(sample_shape=(Ptest, B)).transpose(0, 2)  # test patterns: same as train + noise
    y = torch.from_numpy(np.random.choice([-1, 1], size=(B, P))).float()  # sample test patterns

    # sample new expansion matrix
    J = torch.zeros(M, Nc)
    for i in range(M):
        idxs = np.random.choice(np.arange(Nc), size=K, replace=False)
        J[i, idxs] = 1 / math.sqrt(K) * torch.randn(K)

    for key, c in cmprs.items():
        if key != "bkp":
            # When using nonlinear compression or recurrent inhibition, still use the linear purely feedforward networks as comparisons
            reluCompTmp = False
            recInhTmp = False
            Wtmp = []
        else:
            reluCompTmp = reluComp
            recInhTmp = recInh
            if excOnly:
                c.weight.data = torch.clamp(c.weight, min=0., max=torch.inf)  # make sure weights are positive
            if recInh:
                Wtmp = W.data
                if recInhType != "unconstrained":
                    Wtmp = torch.clamp(Wtmp, min=-torch.inf, max=0.)
            else:
                Wtmp = []

        if key == "DE":
            Jtmp = torch.zeros(M, N)
            for i in range(M):
                idxs = np.random.choice(np.arange(N), size=K, replace=False)
                Jtmp[i, idxs] = 1 / math.sqrt(K) * torch.randn(K)
        else:
            Jtmp = J

        # run network forward, learn hebbian weights and test
        PyHat, _, _, _, _ = runandlearn(c, zTrain, zTest, y, math.sqrt(N / D) * A, Jtmp, math.sqrt(N / Dn) * An, distrZn, f=f, sigmaO=sigmaO, sigmaM=sigmaM, reluComp=reluCompTmp, fC=fC, rec=recInhTmp, W=Wtmp)

        # run once more, just to get the representation (noiseless)
        zn = distrZn.sample(sample_shape=zTrain.shape[1:]).transpose(0, 2).transpose(1, 2)
        x, c0, m, ths = runforward(c, zTrain, math.sqrt(N / D) * A, Jtmp, zn, math.sqrt(N / Dn) * An, f=f, sigmaO=0., sigmaM=0., reluComp=reluCompTmp, fC=fC, rec=recInhTmp, W=Wtmp)

        # run once more, just to get the representation (noisy)
        zn = distrZn.sample(sample_shape=zTest.shape[1:]).transpose(0, 2).transpose(1, 2)
        xTest, cTest, mTest, _ = runforward(c, zTest, math.sqrt(N / D) * A, Jtmp, zn, math.sqrt(N / Dn) * An, f=f, sigmaO=sigmaO, sigmaM=sigmaM, reluComp=reluCompTmp, fC=fC, rec=recInhTmp, W=Wtmp)

        # get performance
        frErrorsAll[key] = 1 / (2 * Ptest) * np.sum(np.abs((PyHat > 0.5).detach().numpy() - ((torch.cat(R * [y], dim=1) + 1) / 2).detach().numpy()), axis=(1))

        # get dimensionality
        Cx = A @ torch.diag(torch.sqrt(torch.pow(torch.arange(1, D + 1), -p))) @ A.transpose(0, 1)
        Cc = c.weight.data @ Cx @ c.weight.data.transpose(0, 1)

        zInf = distrZ.sample(sample_shape=(Pinf, B)).transpose(0, 2)  # lots of patterns to assess dimension
        zn = distrZn.sample(sample_shape=zInf.shape[1:]).transpose(0, 2).transpose(1, 2)
        # run forward again to get the lots of patterns (noiseless) that we need to measure the dimensionality
        xInf, cInf, mInf, thsInf = runforward(c, zInf, math.sqrt(N / D) * A, Jtmp, zn, math.sqrt(N / Dn) * An, f=f, sigmaO=0., sigmaM=0., reluComp=reluCompTmp, fC=fC, rec=recInhTmp, W=Wtmp)

        Ccs = 1 / Pinf * torch.einsum('ibp, jbp -> ijb', cInf - torch.mean(cInf, dim=2)[:, :, None], cInf - torch.mean(cInf, dim=2)[:, :, None])  # one cov matrix for each batch
        Cms = 1 / Pinf * torch.einsum('ibp, jbp -> ijb', mInf - torch.mean(mInf, dim=2)[:, :, None], mInf - torch.mean(mInf, dim=2)[:, :, None])  # one cov matrix for each batch
        dimC = torch.einsum('iib -> b', Ccs)**2 / torch.einsum('ijb, ijb -> b', Ccs, Ccs)
        dimCall[key] = torch.mean(dimC)
        dimM = torch.einsum('iib -> b', Cms)**2 / torch.einsum('ijb, ijb -> b', Cms, Cms)
        dimAll[key] = torch.mean(dimM)

        # get compression layer noise from analytical expression
        noiseAllTh[key] = sigmaP**2 * D * torch.trace(c.weight.data @ c.weight.data.transpose(0, 1)) / (2 * N * torch.trace(Cc))
        # get compression layer noise numerycally
        noiseCall[key] = torch.mean(torch.mean(torch.sum((cTest - torch.cat(10 * [c0], dim=2))**2, dim=0), dim=1) / torch.mean(torch.sum((cInf - cInf[:, :, torch.randperm(Pinf)])**2, dim=0), dim=1))

        # get expansion layer noise numerically
        noiseAll[key] = torch.mean(torch.mean(torch.sum((mTest - torch.cat(10 * [m], dim=2))**2, dim=0), dim=1) / torch.mean(torch.sum((mInf - mInf[:, :, torch.randperm(Pinf)])**2, dim=0), dim=1))

[print(key, np.mean(frerr))for key, frerr in frErrorsAll.items()]
