# Copyright 2023, Samuel Muscinelli.
# Code to produce the main results of Muscinelli, Wagner, and Litwin-Kumar (2023).
#  This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.

import math
import torch
import numpy as np
import torch.nn as nn
from scipy.special import erfcinv


def genhaarorthmatrix(N):
    """
    Generate a random N-dimensional orthogonal matrix from the Haar group.
    This is the function used to generate *distributed* embeddings.
    This is similar to taking a N-by-N Gaussian and orthogonalizing it, but from a slightly different distribution that ensure that the eigenvalues lie uniformly in the unit circle.
    """
    H = np.diag(np.ones(N))
    D = np.ones(N)
    for n in range(N - 1):
        x = np.random.randn(N - n, 1)
        D[n] = np.sign(x[0])
        x[0] = x[0] - D[n] * np.linalg.norm(x)
        Hx = np.diag(np.ones(N - n)) - np.matmul((2 * x), x.transpose()) / (np.linalg.norm(x) ** 2)  # Householder transformation
        mat = np.diag(np.ones(N))
        mat[n:, n:] = Hx[:, :]
        H = np.matmul(H, mat)

    D[-1] = (-1) ** (1 - (N % 2)) * np.prod(D)
    H = np.dot(np.diag(D), H)

    return H


def genclustembeddingmatrix(N, D):
    """
    Returns the matrix used to generate a *clustered* embedding of D task variables in a N-dim. input layer.
    """
    A = torch.zeros(N, D)
    NperBlock = int(math.floor(N / D))
    for b in range(D):
        A[b * NperBlock:(b + 1) * NperBlock, b] = 1. / np.sqrt(NperBlock)  # so that the columns have norm 1
    return A


def gethebbianreadoutweights(m, y, meansubtracted=True):
    """
    Learn the supervised Hebbian weights.
    m: expansion layer representation. Shape: M x B x P
    y: target labels. Shape: B x P
    """
    M = m.shape[0]
    if meansubtracted:
        m = m - torch.mean(m, dim=2)[:, :, None]
    return 1 / (math.sqrt(M)) * torch.einsum('ijk, jk -> ij', m, y)


def embedinput(z, A, nonlinearity=None):
    """
    Embeds the task variables z into the input layer via the embedding matrix A.
    Note that the embedding does not depend on the batch sample.
    z: shape D x B x P
    A: shape N x D
    """
    x = torch.einsum('ij, jbp -> ibp', A, z)
    return x


def init_recurrent_connectivity(Nc, gI, recInhType):
    if recInhType == "learned":
        # inhibitory and learnable
        W = nn.Parameter(-torch.abs(torch.randn(Nc, Nc)))
    elif recInhType == "global":
        # global and fixed
        W = - gI * torch.einsum('i, j -> ij', torch.ones(Nc), torch.ones(Nc))
    elif recInhType == "unconstrained":
        # unconstrained and learnable
        W = nn.Parameter(torch.randn(Nc, Nc))
    else:
        raise NotImplementedError("The type of recurrent connectivity was not recognized!")
    return W


def get_noise(G, Cc, Cn, sigma):
    """
    Uses the analytical expression to get noise strength at the compression layer.
    This is used only when the loss function is the SNR at the compression layer to speed up learning giving an accurate estimate of noise that does not require a lot of samples.
    """
    N = Cn.shape[0]
    return sigma**2 / 2 * 1 / N * torch.sum(torch.diagonal(torch.einsum('ij, jk -> ik', torch.einsum('ij, jk -> ik', G, Cn), G.transpose(0, 1)), dim1=0, dim2=1), 0) / torch.sum(torch.diagonal(Cc, dim1=0, dim2=1), 1)  # factor N is in here, because it is not in the weights!


def runforward(cmprLayer, z, A, J, zn, An, f=0.1, sigmaO=0.1, sigmaM=0., reluComp=False, fC=1., rec=False, W=[]):
    """
    Runs the network forward, meaning that given the latent representation it generates the representation of all network layers.

    cmprLayer: a torch.nn.Linear object that implements the network compression, with N as input size and Nc as output size.
    z: task variable representaion. Shape D x B x P
    A: Embedding matrix, shape N x D
    J: Expansion matrix, shape M x Nc
    zn: noise representation, shape Dn x B x P
    An: Noise embedding matrix, shape N x Dn
    f: coding level of the network (default 0.1)
    sigmaO: Strength of noise (default 0.1)
    sigmaM: Expansion layer noise (default 0.)
    reluComp: whether to apply a ReLU nonlinearity after linear compression (default False)
    fC: coding level of the compression layer. It is used only if reluComp = True
    rec: whether to use recurrent connectivity at the compression layer (using the stationary solution)
    W: compression layer recurrent weight matrix, used only if rec=True. Shape Nc x Nc.
    """

    N = A.shape[0]
    Nc = cmprLayer.out_features

    x = embedinput(z, A)  # embed input
    x = x + sigmaO * embedinput(zn, An)  # embed noise

    if reluComp:
        hS = cmprLayer(1 / np.sqrt(N) * x.transpose(0, 2)).transpose(0, 2)  # the scaling is here so weights can be order 1
        if rec:
            hS = torch.einsum('ij, jbp -> ibp', torch.inverse(torch.eye(cmprLayer.out_features) - W), 1 / np.sqrt(Nc) * hS)  # use the stationary state solution
        thsS = torch.sqrt(2 * torch.var(hS, dim=2)) * erfcinv(2 * fC)  # find the thresholds to give the right coding level
        c = torch.relu(hS - thsS[:, :, None])
        h = torch.einsum('ij, jbp -> ibp', J, c)  # get expansion layer input currents
    else:
        c = cmprLayer(1 / np.sqrt(N) * x.transpose(0, 2)).transpose(0, 2)  # get compression layer representation (the scaling is here so weights can be order 1)
        if rec:
            c = torch.einsum('ij, jbp -> ibp', torch.inverse(torch.eye(cmprLayer.out_features) - 1 / np.sqrt(Nc) * W), c)  # use the stationary state solution
        h = torch.einsum('ij, jbp -> ibp', J, c)  # get expansion layer input currents

    ths = torch.sqrt(2 * torch.var(h, dim=2)) * erfcinv(2 * f)  # find threshold to get the desired coding level at the expansion layer
    m = torch.relu(h - ths[:, :, None])  # apply nonlinearity
    m = m + sigmaM * torch.randn(m.shape[0], m.shape[1], m.shape[2])  # add expansion layer noise
    return x, c, m, ths


def runandlearn(cmprLayer, zTrain, zTest, y, A, J, An, distrZn, f=0.1, sigmaO=0.1, sigmaM=0., reluComp=False, fC=1., rec=False, W=[]):
    """
    Runs the network forward and learns the readout weights.
    cmprLayer: a torch.nn.Linear object that implements the network compression, with N as input size and Nc as output size.
    zTrain: Task variable representation for training. Shape D x B x P
    zTest: Task variable representation for testin. Shape D x B x Ptest
    y: True labels. Shape B x P
    A: Embedding matrix, shape N x D
    J: Expansion matrix, shape M x Nc
    An: Noise embedding matrix, shape N x Dn
    distrZn: Distribution from which the noise is sampled, in Dn dimensions (the i.i.d Gaussian noise case is retrieved by setting it to a multi-D independent Gaussian, set Dn = N and set An = identity matrix in N dimensions).
    f: coding level of the network (default 0.1)
    sigmaO: Strength of noise (default 0.1)
    sigmaM: Expansion layer noise (default 0.)
    reluComp: whether to apply a ReLU nonlinearity after linear compression (default False)
    fC: coding level of the compression layer. It is used only if reluComp = True
    rec: whether to use recurrent connectivity at the compression layer (using the stationary solution)
    W: compression layer recurrent weight matrix, used only if rec=True. Shape Nc x Nc.
    """

    zn = distrZn.sample(sample_shape=zTrain.shape[1:]).transpose(0, 2).transpose(1, 2)  # sample training noise (this is a dummy variable because the noise strength will be set to zero during training)
    xTrain, cTrain, m, ths = runforward(cmprLayer, zTrain, A, J, zn, An, f=f, sigmaO=0., sigmaM=0., reluComp=reluComp, fC=fC, rec=rec, W=W)  # get the training representation for all layers. Importantly, we set the nosie strength equal to zero, i.e. *no noise during training*.

    zn = distrZn.sample(sample_shape=zTest.shape[1:]).transpose(0, 2).transpose(1, 2)  # sample testing noise
    xTest, cTest, mTest, ths = runforward(cmprLayer, zTest, A, J, zn, An, f=f, sigmaO=sigmaO, sigmaM=sigmaM, reluComp=reluComp, fC=fC, rec=rec, W=W)  # get the testing representation for all layers

    w = gethebbianreadoutweights(m, y, meansubtracted=True)  # learn readout weights

    inpyHat = torch.einsum('ib, ibp -> bp', w, mTest - torch.mean(mTest, dim=-1, keepdim=True))  # get the readout

    PyHat = torch.sigmoid(inpyHat)  # transform readout to probabilities (not normalized)

    return PyHat, xTrain, cTrain, m, ths
